import Live
import sys
from ParamMap import ParamMap
from consts import *
#from Tracing import Traced

def tuple_idx(tuple, obj):
    for i in xrange(0,len(tuple)):
        if not cmp(tuple[i], obj):
            return i

#class UC3(Traced):
#noinspection PyCallingNonCallable
class UC3():
    """
    UC3 script main object
    """
    __doc__ = "Faderfox UC3 controller script"
    __version__ = "v1.0"
    __name__ = "Faderfox UC3"
    __module__ = __name__
    __myDebug__ = False
    __filter_funcs__ = ["update_display", "log", "exec_commands", "song", "send_midi", "get_lv3_tracks", "get_lv3_track_from_channel", "get_lv3_all_tracks_channels", "lv3_get_return_tracks"]

    locked_device = None
    locked_device_idx = None
    c_instance = None
    param_map = None
    file = None

    def __init__(self, c_instance):
        UC3.realinit(self, c_instance)

    def realinit(self, c_instance):
        """ Do the real initialization. """
        if self.__myDebug__:
            if sys.platform == "win32":
                self.file = open("C:/ableton-debug.txt", "a")
                self.commandfile = "C:/ableton-debug-cmd.txt"
            else:
                self.file = open("/tmp/ableton-debug", "a")
                self.commandfile = "/tmp/ableton-debug-cmd"

        self.param_map = ParamMap(self)

        self.c_instance = c_instance
        self.show_message(self.__name__ + " " + self.__version__)

        self.song().view.add_selected_track_listener(self.on_selected_track_changed)
        self.song().add_clip_trigger_quantization_listener(self.on_quantization_changed)

        # Register callbacks for the main script handler
        # Instantiate controllers. We have 2 controllers, one for general transport parameters, one
        # for track specific functionality

        self.initializeMidi = True

        self.log("realinit request rebuild")
        self.request_rebuild_midi_map()

    ###########################################################
    #
    # Ableton API methods
    #
    ###########################################################
    def disconnect(self):
        self.song().view.remove_selected_track_listener(self.on_selected_track_changed)
        self.song().remove_clip_trigger_quantization_listener(self.on_quantization_changed)

    def application(self):
        """
        Return the application object
        """
        return Live.Application.get_application()

    def song(self):
        """
        Return the song object
        """
        return self.c_instance.song()

    def suggest_input_port(self):
        """
        Suggest a port name
        """
        return str("Faderfox UC3")

    def suggest_output_port(self):
        """
        Suggest an output port name
        """
        return str("Faderfox UC3")

    def can_lock_to_devices(self):
        """
        This script can lock to devices
        """
        return True


    def set_appointed_device(self, device):
        """
        Set the scripts appointed device
        """
        pass

    def toggle_lock(self):
        """
        Toggle lock
        """
        self.c_instance.toggle_lock()

    def suggest_map_mode(self, cc_no, foobar):
        """
        Suggest the default CC map mode
        """
        return Live.MidiMap.MapMode.absolute


    def restore_bank(self, bank):
        """
        Restore the script bank
        """
        pass

    def show_message(self, message):
        """
        Display a message (compatibility with live 5)
        """
        if hasattr(self.c_instance, 'show_message'):
            self.c_instance.show_message(message)

    def instance_identifier(self):
        """
        Return the C objects instance identifier
        """
        return self.c_instance.instance_identifier()

    def connect_script_instances(self, instantiated_scripts):
        """
        Connect instantiated scripts (when linking scripts)
        """
        pass

    def request_rebuild_midi_map(self):
        """
        Request a rebuild of the midi map
        """
        self.c_instance.request_rebuild_midi_map()

    def send_midi(self, midi_event_bytes):
        """
        Send out midi bytes
        """
        self.log("send midi %s" % (list(midi_event_bytes)))
        self.c_instance.send_midi(midi_event_bytes)

    def refresh_state(self):
        """
        Refresh the script's state.

        Send out scene select and track select CCs to controller.
        """
        pass

    def device_name(self, device):
        if (hasattr(device, "class_name")):
            return device.class_name
        else:
            return ""

    def track_find_rack(self, track, idx):
        """
        Helper method to find the first rack in a track
        """
        for device in track.devices:
            if self.is_rack(device):
                idx -= 1
                if idx == 0:
                    return device

        return None

    def track_get_rack_idx(self, track, device):
        idx = 0
        for _device in track.devices:
            if self.is_rack(_device):
                idx += 1
                if device == _device:
                    return idx
        return None

    def is_rack(self, device):
        """
        Returns true if the device is a rack
        """
        deviceName = self.device_name(device)
        return deviceName in ["DrumGroupDevice",
                      "InstrumentGroupDevice",
                      "MidiEffectGroupDevice",
                      "AudioEffectGroupDevice"]

    QuantizationList = [
        Live.Song.Quantization.q_no_q,
        Live.Song.Quantization.q_8_bars,
        Live.Song.Quantization.q_4_bars,
        Live.Song.Quantization.q_2_bars,
        Live.Song.Quantization.q_bar,
        Live.Song.Quantization.q_half,
        Live.Song.Quantization.q_half_triplet,
        Live.Song.Quantization.q_quarter,
        Live.Song.Quantization.q_quarter_triplet,
        Live.Song.Quantization.q_eight,
        Live.Song.Quantization.q_eight_triplet,
        Live.Song.Quantization.q_sixtenth,
        Live.Song.Quantization.q_sixtenth_triplet,
        Live.Song.Quantization.q_thirtytwoth
    ]

    def get_all_tracks(self):
        return list(self.song().tracks) + list(self.song().return_tracks) + [self.song().master_track]

    def get_track_idx(self, track):
        return tuple_idx(self.get_all_tracks(), track)

    def lock_to_device(self, device):
        """
        Called when the script is locked to a device
        """
        if self.is_rack(device):
            self.log("lock to device %s" % device)
            self.locked_device_idx = self.track_get_rack_idx(self.song().view.selected_track, device)
            if self.locked_device_idx:
                self.locked_device = device
                self.log("track rack idx %s" % self.locked_device_idx)
            else:
                self.log("Could not find locked device in selected track")
        else:
            self.log("cannot lock to non-rack device %s" % device)
            self.show_message("Cannot lock to non-rack device")

    def unlock_from_device(self, device):
        """
        Called when the script is unlocked
        """
        self.log("unlock from device %s, locked_device %s" % (device, self.locked_device))
        self.request_rebuild_midi_map()
        self.locked_device_idx = None
        self.locked_device = None

    def on_quantization_changed(self):
        # send midi feedback ? XXX
        self.log("quantization changed %s" % self.song().clip_trigger_quantization)
        idx = tuple_idx(self.QuantizationList, self.song().clip_trigger_quantization)
        self.log("quantization index %s" % idx)
        for channel in [MASTER_CHANNEL_1, MASTER_CHANNEL_2]:
            self.send_midi((CC_STATUS | channel, CC_QUANTIZATION, idx))

    def on_selected_track_changed(self):
        idx = self.get_track_idx(self.song().view.selected_track)
        self.log("track changed idx: %s" % idx)
        for channel in [MASTER_CHANNEL_1, MASTER_CHANNEL_2]:
            self.send_midi((CC_STATUS | channel, CC_TRACK_SELECT, idx + 1))
        self.request_rebuild_midi_map()

    def receive_midi_cc(self, channel, cc_no, cc_val):
        if channel == MASTER_CHANNEL_1 or channel == MASTER_CHANNEL_2:
            if cc_no == CC_QUANTIZATION:
                if cc_val >= len(self.QuantizationList):
                    cc_val = len(self.QuantizationList) - 1
                self.song().clip_trigger_quantization = self.QuantizationList[cc_val]
            elif cc_no == CC_TRACK_SELECT:
                if cc_val == 0:
                    cc_val = 1
                    for channel in [MASTER_CHANNEL_1, MASTER_CHANNEL_2]:
                        self.send_midi((CC_STATUS | channel, CC_TRACK_SELECT, 1))
                cc_val -= 1
                tracks = self.get_all_tracks()
                if cc_val < len(tracks):
                    self.song().view.selected_track = tracks[cc_val]
                else:
                    self.song().view.selected_track = tracks[len(tracks)-1]
                    for channel in [MASTER_CHANNEL_1, MASTER_CHANNEL_2]:
                        self.send_midi((CC_STATUS | channel, CC_TRACK_SELECT, len(tracks)))


    def build_midi_map(self, midi_map_handle):
        """
        Build a midi map (calls all registered components)
        """
        self.log("rebuild midi map")
        script_handle = self.c_instance.handle()
        self.param_map.remove_mappings()

        track = self.song().view.selected_track
        self.log("selected track: %s" % track)
        if track:
            for idx in range(1, 4):
                if idx == self.locked_device_idx:
                    self.log("skipping locked device idx %s" % idx)
                    rack = self.locked_device
                else:
                    rack = self.track_find_rack(track, idx)
                if rack:
                    cc_no = idx * 8
                    self.log("mapping rack %s to cc_no %s" % (idx, cc_no))
                    for parameter in rack.parameters[1:9]:
                        self.log("mapping param %s for rack %s to cc %s" % (parameter.name, idx, cc_no))
                        ParamMap.map_with_feedback(midi_map_handle, MASTER_CHANNEL_1,
                                                   cc_no, parameter, Live.MidiMap.MapMode.absolute)
                        ParamMap.map_with_feedback(midi_map_handle, MASTER_CHANNEL_2,
                                                   cc_no, parameter, Live.MidiMap.MapMode.absolute)
                        cc_no += 1

        # volumes
        tracks = list(self.song().tracks) + list(self.song().return_tracks)
        idx = 0
        for track in tracks[0 : 16]:
            mixer_device = track.mixer_device

            param_idx = idx
            if idx < 8:
                channel = MASTER_CHANNEL_1
            else:
                channel = MASTER_CHANNEL_2
                param_idx -= 8

            # track volumes
            ParamMap.map_with_feedback(midi_map_handle, channel, CC_FADER_START + param_idx,
                                       mixer_device.volume, Live.MidiMap.MapMode.absolute)
            sends = mixer_device.sends
            if len(sends) >= 1:
                ParamMap.map_with_feedback(midi_map_handle, channel, CC_SEND1_START + param_idx,
                                           sends[0], Live.MidiMap.MapMode.absolute)
            if len(sends) >= 2:
                ParamMap.map_with_feedback(midi_map_handle, channel, CC_SEND2_START + param_idx,
                                           sends[1], Live.MidiMap.MapMode.absolute)
            if len(sends) >= 3:
                ParamMap.map_with_feedback(midi_map_handle, channel, CC_SEND3_START + param_idx,
                                           sends[2], Live.MidiMap.MapMode.absolute)
            ParamMap.map_with_feedback(midi_map_handle, channel, CC_PAN_START + param_idx,
                                       mixer_device.panning, Live.MidiMap.MapMode.absolute)
            idx += 1

        master_track = self.song().master_track
        ParamMap.map_with_feedback(midi_map_handle, MASTER_CHANNEL_1, CC_CROSSFADER, master_track.mixer_device.crossfader, Live.MidiMap.MapMode.absolute)

        for channel in [MASTER_CHANNEL_1, MASTER_CHANNEL_2]:
            ParamMap.map_with_feedback(midi_map_handle, channel, CC_MASTER_PAN, master_track.mixer_device.panning,
                                       Live.MidiMap.MapMode.absolute)
            ParamMap.map_with_feedback(midi_map_handle, channel, CC_MASTER_VOLUME, master_track.mixer_device.volume,
                                       Live.MidiMap.MapMode.absolute)
            if hasattr(master_track.mixer_device, "cue_volume"):
                ParamMap.map_with_feedback(midi_map_handle, channel, CC_PREVIEW_CUE, master_track.mixer_device.cue_volume,
                                           Live.MidiMap.MapMode.absolute)
            Live.MidiMap.forward_midi_cc(script_handle, midi_map_handle, channel, CC_QUANTIZATION)
            Live.MidiMap.forward_midi_cc(script_handle, midi_map_handle, channel, CC_TRACK_SELECT)

        self.log("end rebuild midi map")

    def update_display(self):
        """
        Update the display.

        Send out the refreshed track status to the controller
        """
        if self.initializeMidi:
            self.initializeMidi = False

        self.exec_commands()

    def receive_midi(self, midi_bytes):
        """
        MIDI receive callback

        Dispatches CC and notes to the registered components
        """
        channel = (midi_bytes[0] & CHAN_MASK)
        status = (midi_bytes[0] & STATUS_MASK)
        if status == CC_STATUS:
            cc_no = midi_bytes[1]
            cc_value = midi_bytes[2]
            self.param_map.receive_midi_cc(channel, cc_no, cc_value)
            self.receive_midi_cc(channel, cc_no, cc_value)
        elif (status == NOTEON_STATUS) or (status == NOTEOFF_STATUS):
            note_no = midi_bytes[1]
            note_vel = midi_bytes[2]
            self.param_map.receive_midi_note(channel, status, note_no, note_vel)
        else:
            pass

    #####################################################################
    #
    # Debug helpers
    #
    #####################################################################
    def exec_commands(self):
        """ Check if new commands have been written to the command file, parse them and execute them."""
        if not self.__myDebug__:
            return

        # read command from command file
        file = open(self.commandfile, "r")
        commands = file.readlines()
        file.close()

        # execute commands
        for command in commands:
            command = command.strip()
            try:
                result = eval(command).__str__()
                #                if (isinstance(result, unicode)):
                #                    result = result.encode("latin-1")
                self.log("%s = %s" % (command, result))
            except Exception, inst:
                self.log("exception: %s" % inst)
                self.log("while executing %s" % command)

            # clear command file
            file = open(self.commandfile, "w")
            file.write("")
            file.close()

    def log(self, string):
        """ Write a string to the debug log file. """
        if self.__myDebug__:
        #            if (isinstance(string, unicode)):
        #                self.file.write(string.encode("latin-1") + "\n")
        #            else:
            self.c_instance.log_message(string)
            self.file.write(string + "\n")
            self.file.flush()
